"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bcrypt_1 = __importDefault(require("bcrypt"));
const api_1 = require("../../../api");
const mongodb_1 = require("mongodb");
const schema_1 = require("./schema");
const utility_1 = require("../utility");
const defaults_1 = require("./defaults");
const cursor_1 = require("./cursor");
class MongoDBUserAdapter {
    constructor(db, bcryptHashCostFactor, locale) {
        this.users = db.collection(schema_1.CollectionName.Users);
        this.bcryptHashCostFactor = bcryptHashCostFactor;
        this.locale = locale;
    }
    async createUser({ input, password }) {
        try {
            const passwordHash = await bcrypt_1.default.hash(password, this.bcryptHashCostFactor);
            const { insertedId: id } = await this.users.insertOne({
                createdAt: new Date(),
                modifiedAt: new Date(),
                email: input.email,
                name: input.name,
                roleIDs: input.roleIDs,
                password: passwordHash
            });
            return this.getUserByID(id);
        }
        catch (err) {
            if (err instanceof mongodb_1.MongoError && err.code === utility_1.MongoErrorCode.DuplicateKey) {
                throw new Error('Email address already exists!');
            }
            throw err;
        }
    }
    async getUser(email) {
        const user = await this.users.findOne({ email });
        if (user) {
            return {
                id: user._id,
                email: user.email,
                name: user.name,
                roleIDs: user.roleIDs
            };
        }
        else {
            return null;
        }
    }
    async updateUser({ id, input }) {
        const { value } = await this.users.findOneAndUpdate({ _id: id }, {
            $set: {
                modifiedAt: new Date(),
                name: input.name,
                email: input.email,
                roleIDs: input.roleIDs
            }
        }, { returnOriginal: false });
        if (!value)
            return null;
        const { _id: outID } = value;
        return this.getUserByID(outID);
    }
    async deleteUser({ id }) {
        const { deletedCount } = await this.users.deleteOne({ _id: id });
        return deletedCount !== 0 ? id : null;
    }
    async resetUserPassword({ id, password }) {
        const { value } = await this.users.findOneAndUpdate({ _id: id }, {
            $set: {
                modifiedAt: new Date(),
                password: await bcrypt_1.default.hash(password, this.bcryptHashCostFactor)
            }
        }, { returnOriginal: false });
        if (!value)
            return null;
        const { _id: outID } = value;
        return this.getUserByID(outID);
    }
    async getUsersByID(ids) {
        const users = await this.users.find({ _id: { $in: ids } }).toArray();
        return users.map(user => {
            return {
                id: user._id,
                email: user.email,
                name: user.name,
                roleIDs: user.roleIDs
            };
        });
    }
    async getUserForCredentials({ email, password }) {
        const user = await this.users.findOne({ email });
        if (user && (await bcrypt_1.default.compare(password, user.password))) {
            return {
                id: user._id,
                email: user.email,
                name: user.name,
                roleIDs: user.roleIDs
            };
        }
        return null;
    }
    async getUserByID(id) {
        const user = await this.users.findOne({ _id: id });
        if (user) {
            return {
                id: user._id,
                email: user.email,
                name: user.name,
                roleIDs: user.roleIDs
            };
        }
        else {
            return null;
        }
    }
    async getUsers({ filter, sort, order, cursor, limit }) {
        const limitCount = Math.min(limit.count, defaults_1.MaxResultsPerPage);
        const sortDirection = limit.type === api_1.LimitType.First ? order : -order;
        const cursorData = cursor.type !== api_1.InputCursorType.None ? cursor_1.Cursor.from(cursor.data) : undefined;
        const expr = order === api_1.SortOrder.Ascending
            ? cursor.type === api_1.InputCursorType.After
                ? '$gt'
                : '$lt'
            : cursor.type === api_1.InputCursorType.After
                ? '$lt'
                : '$gt';
        const sortField = userSortFieldForSort(sort);
        const cursorFilter = cursorData
            ? {
                $or: [
                    { [sortField]: { [expr]: cursorData.date } },
                    { _id: { [expr]: cursorData.id }, [sortField]: cursorData.date }
                ]
            }
            : {};
        let textFilter = {};
        // TODO: Rename to search
        if ((filter === null || filter === void 0 ? void 0 : filter.name) != undefined) {
            textFilter['$or'] = [{ name: { $regex: filter.name, $options: 'i' } }];
        }
        const [totalCount, users] = await Promise.all([
            this.users.countDocuments(textFilter, {
                collation: { locale: this.locale, strength: 2 }
            }),
            this.users
                .aggregate([], { collation: { locale: this.locale, strength: 2 } })
                .match(textFilter)
                .match(cursorFilter)
                .sort({ [sortField]: sortDirection, _id: sortDirection })
                .limit(limitCount + 1)
                .toArray()
        ]);
        const nodes = users.slice(0, limitCount);
        if (limit.type === api_1.LimitType.Last) {
            nodes.reverse();
        }
        const hasNextPage = limit.type === api_1.LimitType.First
            ? users.length > limitCount
            : cursor.type === api_1.InputCursorType.Before
                ? true
                : false;
        const hasPreviousPage = limit.type === api_1.LimitType.Last
            ? users.length > limitCount
            : cursor.type === api_1.InputCursorType.After
                ? true
                : false;
        const firstUser = nodes[0];
        const lastUser = nodes[nodes.length - 1];
        const startCursor = firstUser
            ? new cursor_1.Cursor(firstUser._id, userDateForSort(firstUser, sort)).toString()
            : null;
        const endCursor = lastUser
            ? new cursor_1.Cursor(lastUser._id, userDateForSort(lastUser, sort)).toString()
            : null;
        return {
            nodes: nodes.map((_a) => {
                var { _id: id } = _a, user = __rest(_a, ["_id"]);
                return (Object.assign({ id }, user));
            }),
            pageInfo: {
                startCursor,
                endCursor,
                hasNextPage,
                hasPreviousPage
            },
            totalCount
        };
    }
}
exports.MongoDBUserAdapter = MongoDBUserAdapter;
function userSortFieldForSort(sort) {
    switch (sort) {
        case api_1.UserSort.CreatedAt:
            return 'createdAt';
        case api_1.UserSort.ModifiedAt:
            return 'modifiedAt';
    }
}
function userDateForSort(user, sort) {
    switch (sort) {
        case api_1.UserSort.CreatedAt:
            return user.createdAt;
        case api_1.UserSort.ModifiedAt:
            return user.modifiedAt;
    }
}
//# sourceMappingURL=user.js.map